"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeclarationReflection = void 0;
const types_1 = require("../types");
const abstract_1 = require("./abstract");
const container_1 = require("./container");
/**
 * A reflection that represents a single declaration emitted by the TypeScript compiler.
 *
 * All parts of a project are represented by DeclarationReflection instances. The actual
 * kind of a reflection is stored in its ´kind´ member.
 */
class DeclarationReflection extends container_1.ContainerReflection {
    hasGetterOrSetter() {
        return !!this.getSignature || !!this.setSignature;
    }
    getAllSignatures() {
        let result = [];
        if (this.signatures) {
            result = result.concat(this.signatures);
        }
        if (this.indexSignature) {
            result.push(this.indexSignature);
        }
        if (this.getSignature) {
            result.push(this.getSignature);
        }
        if (this.setSignature) {
            result.push(this.setSignature);
        }
        return result;
    }
    /** @internal */
    getNonIndexSignatures() {
        return [].concat(this.signatures ?? [], this.setSignature ?? [], this.getSignature ?? []);
    }
    /**
     * Traverse all potential child reflections of this reflection.
     *
     * The given callback will be invoked for all children, signatures and type parameters
     * attached to this reflection.
     *
     * @param callback  The callback function that should be applied for each child reflection.
     */
    traverse(callback) {
        for (const parameter of this.typeParameters?.slice() || []) {
            if (callback(parameter, abstract_1.TraverseProperty.TypeParameter) === false) {
                return;
            }
        }
        if (this.type instanceof types_1.ReflectionType) {
            if (callback(this.type.declaration, abstract_1.TraverseProperty.TypeLiteral) === false) {
                return;
            }
        }
        for (const signature of this.signatures?.slice() || []) {
            if (callback(signature, abstract_1.TraverseProperty.Signatures) === false) {
                return;
            }
        }
        if (this.indexSignature) {
            if (callback(this.indexSignature, abstract_1.TraverseProperty.IndexSignature) === false) {
                return;
            }
        }
        if (this.getSignature) {
            if (callback(this.getSignature, abstract_1.TraverseProperty.GetSignature) ===
                false) {
                return;
            }
        }
        if (this.setSignature) {
            if (callback(this.setSignature, abstract_1.TraverseProperty.SetSignature) ===
                false) {
                return;
            }
        }
        super.traverse(callback);
    }
    /**
     * Return a string representation of this reflection.
     */
    toString() {
        let result = super.toString();
        if (this.typeParameters) {
            const parameters = this.typeParameters.map((parameter) => parameter.name);
            result += "<" + parameters.join(", ") + ">";
        }
        if (this.type) {
            result += ":" + this.type.toString();
        }
        return result;
    }
    toObject(serializer) {
        return {
            ...super.toObject(serializer),
            typeParameter: this.typeParameters && this.typeParameters.length > 0
                ? this.typeParameters.map((type) => serializer.toObject(type))
                : undefined,
            type: this.type && serializer.toObject(this.type),
            signatures: this.signatures?.map((sig) => serializer.toObject(sig)),
            indexSignature: this.indexSignature && serializer.toObject(this.indexSignature),
            getSignature: this.getSignature && [
                serializer.toObject(this.getSignature),
            ],
            setSignature: this.setSignature && [
                serializer.toObject(this.setSignature),
            ],
            defaultValue: this.defaultValue,
            overwrites: this.overwrites && serializer.toObject(this.overwrites),
            inheritedFrom: this.inheritedFrom && serializer.toObject(this.inheritedFrom),
            implementationOf: this.implementationOf &&
                serializer.toObject(this.implementationOf),
            extendedTypes: this.extendedTypes?.map((type) => serializer.toObject(type)),
            extendedBy: this.extendedBy?.map((type) => serializer.toObject(type)),
            implementedTypes: this.implementedTypes?.map((type) => serializer.toObject(type)),
            implementedBy: this.implementedBy?.map((type) => serializer.toObject(type)),
        };
    }
}
exports.DeclarationReflection = DeclarationReflection;
